/*
 * Copyright (c) 2018 MariaDB Corporation Ab
 * Copyright (c) 2023 MariaDB plc, Finnish Branch
 *
 * Use of this software is governed by the Business Source License included
 * in the LICENSE.TXT file and at www.mariadb.com/bsl11.
 *
 * Change Date: 2026-09-21
 *
 * On the date above, in accordance with the Business Source License, use
 * of this software will be governed by version 2 or later of the General
 * Public License.
 */
#pragma once

/**
 * @file Readwritesplit common header
 */

#define MXB_MODULE_NAME "readwritesplit"

#include <maxscale/ccdefs.hh>

#include <maxbase/shared_mutex.hh>
#include <maxbase/small_vector.hh>
#include <maxscale/protocol/mariadb/module_names.hh>
#include <maxscale/protocol/mariadb/mysql.hh>
#include <maxscale/protocol/mariadb/rwbackend.hh>
#include <maxscale/queryclassifier.hh>
#include <maxscale/router.hh>
#include <maxscale/session_stats.hh>
#include <maxscale/workerlocal.hh>
#include <maxscale/modulecmd.hh>

namespace cfg = maxscale::config;
using namespace std::literals::chrono_literals;

constexpr int SLAVE_MAX = 255;

enum backend_type_t
{
    BE_UNDEFINED = -1,
    BE_MASTER,
    BE_JOINED = BE_MASTER,
    BE_SLAVE,
    BE_COUNT
};

enum connection_type
{
    ALL,
    SLAVE
};

typedef uint32_t route_target_t;

/**
 * This criteria is used when backends are chosen for a router session's use.
 * Backend servers are sorted to ascending order according to the criteria
 * and top N are chosen.
 */
enum select_criteria_t
{
    LEAST_GLOBAL_CONNECTIONS,   /**< all connections established by MaxScale */
    LEAST_ROUTER_CONNECTIONS,   /**< connections established by this router */
    LEAST_BEHIND_MASTER,
    LEAST_CURRENT_OPERATIONS,
    ADAPTIVE_ROUTING
};

/**
 * Controls how master failure is handled
 */
enum failure_mode
{
    RW_FAIL_INSTANTLY,          /**< Close the connection as soon as the master is lost */
    RW_FAIL_ON_WRITE,           /**< Close the connection when the first write is received */
    RW_ERROR_ON_WRITE           /**< Don't close the connection but send an error for writes */
};

enum class CausalReads
{
    NONE,           // No causal reads, default
    LOCAL,          // Causal reads are done on a session level with MASTER_GTID_WAIT
    GLOBAL,         // Causal reads are done globally with MASTER_GTID_WAIT
    FAST,           // Causal reads use GTID position to pick an up-to-date server
    FAST_GLOBAL,    // Same as FAST except the global GTID is used
    UNIVERSAL,      // Causal reads verifies the GTID before starting the read
    FAST_UNIVERSAL  // Causal reads verifies the GTID and then behaves like FAST
};

enum class TrxChecksum
{
    FULL,           // Checksum all responses
    RESULT_ONLY,    // Checksum resultsets and errors, ignore OK packets
    NO_INSERT_ID,   // Same as RESULT_ONLY but also ignores results from queries that use LAST_INSERT_ID()
};

static cfg::Specification s_spec(MXB_MODULE_NAME, cfg::Specification::ROUTER);

static cfg::ParamEnum<mxs_target_t> s_use_sql_variables_in(
    &s_spec, "use_sql_variables_in",
    "Whether to route SQL variable modifications to all servers or only to the master",
{
    {TYPE_ALL, "all"},
    {TYPE_MASTER, "master"},
}, TYPE_ALL, cfg::Param::AT_RUNTIME);

static cfg::ParamEnum<select_criteria_t> s_slave_selection_criteria(
    &s_spec, "slave_selection_criteria", "Slave selection criteria",
{
    {LEAST_GLOBAL_CONNECTIONS, "least_global_connections"},
    {LEAST_ROUTER_CONNECTIONS, "least_router_connections"},
    {LEAST_BEHIND_MASTER, "least_behind_master"},
    {LEAST_CURRENT_OPERATIONS, "least_current_operations"},
    {ADAPTIVE_ROUTING, "adaptive_routing"},
    {LEAST_GLOBAL_CONNECTIONS, "LEAST_GLOBAL_CONNECTIONS"},
    {LEAST_ROUTER_CONNECTIONS, "LEAST_ROUTER_CONNECTIONS"},
    {LEAST_BEHIND_MASTER, "LEAST_BEHIND_MASTER"},
    {LEAST_CURRENT_OPERATIONS, "LEAST_CURRENT_OPERATIONS"},
    {ADAPTIVE_ROUTING, "ADAPTIVE_ROUTING"},
}, LEAST_CURRENT_OPERATIONS, cfg::Param::AT_RUNTIME);

static cfg::ParamEnum<failure_mode> s_master_failure_mode(
    &s_spec, "master_failure_mode", "Master failure mode behavior",
{
    {RW_FAIL_INSTANTLY, "fail_instantly"},
    {RW_FAIL_ON_WRITE, "fail_on_write"},
    {RW_ERROR_ON_WRITE, "error_on_write"}
}, RW_FAIL_INSTANTLY, cfg::Param::AT_RUNTIME);

static cfg::ParamEnum<CausalReads> s_causal_reads(
    &s_spec, "causal_reads", "Causal reads mode",
{
    {CausalReads::NONE, "none"},
    {CausalReads::LOCAL, "local"},
    {CausalReads::GLOBAL, "global"},
    {CausalReads::FAST_GLOBAL, "fast_global"},
    {CausalReads::FAST, "fast"},
    {CausalReads::UNIVERSAL, "universal"},
    {CausalReads::FAST_UNIVERSAL, "fast_universal"},

    // Legacy values for causal_reads
    {CausalReads::NONE, "false"},
    {CausalReads::NONE, "off"},
    {CausalReads::NONE, "0"},
    {CausalReads::LOCAL, "true"},
    {CausalReads::LOCAL, "on"},
    {CausalReads::LOCAL, "1"},
}, CausalReads::NONE, cfg::Param::AT_RUNTIME);

static cfg::ParamSeconds s_max_replication_lag(
    &s_spec, "max_replication_lag", "Maximum replication lag",
    std::chrono::seconds(0),
    cfg::Param::AT_RUNTIME);

static cfg::ParamDeprecated<cfg::ParamAlias> s_max_slave_replication_lag(
    &s_spec, "max_slave_replication_lag", &s_max_replication_lag);

static cfg::ParamCount s_max_slave_connections(
    &s_spec, "max_slave_connections", "Maximum number of slave connections",
    SLAVE_MAX, cfg::Param::AT_RUNTIME);

static cfg::ParamCount s_slave_connections(
    &s_spec, "slave_connections", "Starting number of slave connections",
    SLAVE_MAX, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_retry_failed_reads(
    &s_spec, "retry_failed_reads", "Automatically retry failed reads outside of transactions",
    true, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_strict_multi_stmt(
    &s_spec, "strict_multi_stmt", "Lock connection to master after multi-statement query",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_strict_sp_calls(
    &s_spec, "strict_sp_calls", "Lock connection to master after a stored procedure is executed",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_strict_tmp_tables(
    &s_spec, "strict_tmp_tables", "Prevent reconnections if temporary tables exist",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_master_accept_reads(
    &s_spec, "master_accept_reads", "Use master for reads",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamSeconds s_causal_reads_timeout(
    &s_spec, "causal_reads_timeout", "Timeout for the slave synchronization",
    10s, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_master_reconnection(
    &s_spec, "master_reconnection", "Reconnect to master",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_delayed_retry(
    &s_spec, "delayed_retry", "Retry failed writes outside of transactions",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamSeconds s_delayed_retry_timeout(
    &s_spec, "delayed_retry_timeout", "Timeout for delayed_retry",
    10s, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_transaction_replay(
    &s_spec, "transaction_replay", "Retry failed transactions",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamSize s_transaction_replay_max_size(
    &s_spec, "transaction_replay_max_size", "Maximum size of transaction to retry",
    1024 * 1024, cfg::Param::AT_RUNTIME);

static cfg::ParamSeconds s_transaction_replay_timeout(
    &s_spec, "transaction_replay_timeout", "Timeout for transaction replay",
    0s, cfg::Param::AT_RUNTIME);

static cfg::ParamCount s_transaction_replay_attempts(
    &s_spec, "transaction_replay_attempts", "Maximum number of times to retry a transaction",
    5, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_transaction_replay_retry_on_deadlock(
    &s_spec, "transaction_replay_retry_on_deadlock", "Retry transaction on deadlock",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_transaction_replay_retry_on_mismatch(
    &s_spec, "transaction_replay_retry_on_mismatch", "Retry transaction on checksum mismatch",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_transaction_replay_safe_commit(
    &s_spec, "transaction_replay_safe_commit", "Prevent replaying of about-to-commit transaction",
    true, cfg::Param::AT_RUNTIME);

static cfg::ParamEnum<TrxChecksum> s_transaction_replay_checksum(
    &s_spec, "transaction_replay_checksum", "Type of checksum to calculate for results",
{
    {TrxChecksum::FULL, "full"},
    {TrxChecksum::RESULT_ONLY, "result_only"},
    {TrxChecksum::NO_INSERT_ID, "no_insert_id"},
}, TrxChecksum::FULL, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_optimistic_trx(
    &s_spec, "optimistic_trx", "Optimistically offload transactions to slaves",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_lazy_connect(
    &s_spec, "lazy_connect", "Create connections only when needed",
    false, cfg::Param::AT_RUNTIME);

static cfg::ParamBool s_reuse_ps(
    &s_spec, "reuse_prepared_statements", "Reuse identical prepared statements inside the same connection",
    false, cfg::Param::AT_RUNTIME);

// The server selection requires some work memory for the candidate selection. The most common case is that
// there is only a handful of candidates to pick from and allocating space for them on the heap is somewhat
// wasteful. Currently 6 values are stored internally which should cover the most common cases.
using Candidates = mxb::small_vector<mxs::RWBackend*, 6>;

/** Function that returns a "score" for a server to enable comparison.
 *  Smaller numbers are better.
 */
using BackendSelectFunction = mxs::RWBackend * (*)(const Candidates& sBackends);
using std::chrono::seconds;

struct RWSConfig : public mxs::config::Configuration
{
    RWSConfig(SERVICE* service);

    struct Values
    {
        using seconds = std::chrono::seconds;

        select_criteria_t     slave_selection_criteria; /**< The slave selection criteria */
        BackendSelectFunction backend_select_fct;

        mxs_target_t use_sql_variables_in;      /**< Whether to send user variables to master or all nodes */
        failure_mode master_failure_mode;       /**< Master server failure handling mode */
        seconds      max_replication_lag;       /**< Maximum replication lag */
        bool         master_accept_reads;       /**< Use master for reads */
        bool         strict_multi_stmt;         /**< Force non-multistatement queries to be routed to the
                                                 * master after a multistatement query. */
        bool        strict_sp_calls;            /**< Lock session to master after an SP call */
        bool        strict_tmp_tables;          /**< Prevent reconnections if temporary tables exist */
        bool        retry_failed_reads;         /**< Retry failed reads on other servers */
        int64_t     max_slave_connections = 0;  /**< Maximum number of slaves for each connection*/
        int64_t     slave_connections;          /**< Minimum number of slaves for each connection*/
        bool        master_reconnection;        /**< Allow changes in master server */
        bool        optimistic_trx;             /**< Enable optimistic transactions */
        bool        lazy_connect;               /**< Create connections only when needed */
        CausalReads causal_reads;
        seconds     causal_reads_timeout;   /**< Timeout, second parameter of function master_wait_gtid */
        bool        delayed_retry;          /**< Delay routing if no target found */
        seconds     delayed_retry_timeout;  /**< How long to delay until an error is returned */

        bool        transaction_replay;     /**< Replay failed transactions */
        int64_t     trx_max_size;           /**< Max transaction size for replaying */
        seconds     trx_timeout;            /**< How long can the transaction be replayed for */
        int64_t     trx_max_attempts;       /**< Maximum number of transaction replay attempts */
        bool        trx_retry_on_deadlock;  /**< Replay the transaction if it ends up in a deadlock */
        bool        trx_retry_on_mismatch;  /**< Replay the transaction on checksum mismatch */
        bool        trx_retry_safe_commit;  /**< Prevent replay of COMMITs */
        TrxChecksum trx_checksum;           /**< The type of checksum to calculate */

        bool reuse_ps;      /**< Reuse prepared statements */
    };

    const mxs::WorkerGlobal<Values>& values() const
    {
        return m_values;
    }

private:
    Values                    m_v;  // Master copy of the values
    mxs::WorkerGlobal<Values> m_values;
    SERVICE*                  m_service;

    bool post_configure(const std::map<std::string, mxs::ConfigParameters>& nested_params) override;

    static BackendSelectFunction get_backend_select_function(select_criteria_t sc);
};

/**
 * The statistics for this router instance
 */
struct Stats
{
    uint64_t n_sessions = 0;        /**< Number sessions created */
    uint64_t n_queries = 0;         /**< Number of queries forwarded */
    uint64_t n_master = 0;          /**< Number of stmts sent to master */
    uint64_t n_slave = 0;           /**< Number of stmts sent to slave */
    uint64_t n_all = 0;             /**< Number of stmts sent to all */
    uint64_t n_trx_replay = 0;      /**< Number of replayed transactions */
    uint64_t n_trx_too_big = 0;     /**< Number of times that transaction_replay_max_size was exceeded */
    uint64_t n_ro_trx = 0;          /**< Read-only transaction count */
    uint64_t n_rw_trx = 0;          /**< Read-write transaction count */
    uint64_t n_ps_reused = 0;       /**< Number of prepared statements that were reused */
    uint64_t n_max_sescmd_sz = 0;   /**< Max m_sescmd_list.size() of all sessions */
};

using maxscale::SessionStats;
using maxscale::TargetSessionStats;

class RWSplitSession;

/**
 * The per instance data for the router.
 */
class RWSplit : public mxs::Router
{
    RWSplit(const RWSplit&);
    RWSplit& operator=(const RWSplit&);

public:
    struct gtid
    {
        uint32_t domain {0};
        uint32_t server_id {0};
        uint64_t sequence {0};

        static gtid from_string(std::string_view str);
        void        parse(std::string_view str);
        std::string to_string() const;
        bool        empty() const;
    };

    RWSplit(SERVICE* service);
    ~RWSplit();

    SERVICE*                    service() const;
    Stats&                      stats();
    const Stats&                stats() const;
    TargetSessionStats&         local_server_stats();
    TargetSessionStats          all_server_stats() const;
    maxbase::CumulativeAverage& local_avg_sescmd_sz();
    int64_t                     avg_sescmd_sz() const;
    std::string                 last_gtid() const;
    std::map<uint32_t, gtid>    last_gtid_map() const;
    void                        set_last_gtid(std::string_view str);
    static bool                 reset_last_gtid(const MODULECMD_ARG* argv, json_t** output);

    const mxs::WorkerGlobal<RWSConfig::Values>& config() const;

    // API functions

    /**
     * @brief Create a new readwritesplit router instance
     *
     * An instance of the router is required for each service that uses this router.
     * One instance of the router will handle multiple router sessions.
     *
     * @param service The service this router is being create for
     * @param options The options for this query router
     *
     * @return New router instance or nullptr on error
     */
    static RWSplit* create(SERVICE* pService);

    /**
     * @brief Create a new session for this router instance
     *
     * The session is used to store all the data required by the router for a
     * particular client connection. The instance of the router that relates to a
     * particular service is passed as the first parameter. The second parameter is
     * the session that has been created in response to the request from a client
     * for a connection. The passed session contains generic information; this
     * function creates the session structure that holds router specific data.
     * There is often a one to one relationship between sessions and router
     * sessions, although it is possible to create configurations where a
     * connection is handled by multiple routers, one after another.
     *
     * @param session  The MaxScale session (generic connection data)
     *
     * @return New router session or nullptr on error
     */
    mxs::RouterSession* newSession(MXS_SESSION* pSession, const Endpoints& endpoints) override;

    /**
     * @brief JSON diagnostics routine
     *
     * @return The JSON representation of this router instance
     */
    json_t* diagnostics() const override;

    /**
     * @brief Get router capabilities
     */
    uint64_t getCapabilities() const override;

    mxs::config::Configuration& getConfiguration() override
    {
        return m_config;
    }

    std::set<std::string> protocols() const override
    {
        return {MXS_MARIADB_PROTOCOL_NAME};
    }

    void update_max_sescmd_sz(uint64_t maybe_max);

private:
    bool check_causal_reads(SERVER* server) const;
    void set_warnings(json_t* json) const;

    SERVICE*                                     m_service; /**< Service where the router belongs*/
    RWSConfig                                    m_config;
    Stats                                        m_stats;
    mxs::WorkerLocal<TargetSessionStats>         m_server_stats;
    mxs::WorkerLocal<maxbase::CumulativeAverage> m_avg_sescmd_sz;
    std::map<uint32_t, gtid>                     m_last_gtid;
    mutable mxb::shared_mutex                    m_last_gtid_lock;
};
