/*
 * Copyright (c) 2016 MariaDB Corporation Ab
 * Copyright (c) 2023 MariaDB plc, Finnish Branch
 *
 * Use of this software is governed by the Business Source License included
 * in the LICENSE.TXT file and at www.mariadb.com/bsl11.
 *
 * Change Date: 2026-09-21
 *
 * On the date above, in accordance with the Business Source License, use
 * of this software will be governed by version 2 or later of the General
 * Public License.
 */

#include "rwsplitsession.hh"

#include <utility>

#include <mysqld_error.h>

#include <maxbase/format.hh>

using std::chrono::seconds;
using maxscale::Parser;
using maxscale::RWBackend;
using mariadb::QueryClassifier;
using RouteInfo = QueryClassifier::RouteInfo;

/**
 * The functions that support the routing of queries to back end
 * servers. All the functions in this module are internal to the read
 * write split router, and not intended to be called from anywhere else.
 */

bool RWSplitSession::prepare_connection(RWBackend* target)
{
    mxb_assert(!target->in_use());
    bool rval = target->connect();

    if (rval)
    {
        MXB_INFO("Connected to '%s'", target->name());
        mxb_assert(!target->is_waiting_result());

        if (m_set_trx && target == m_current_master)
        {
            MXB_INFO("Re-executing SET TRANSACTION: %s", get_sql_string(m_set_trx).c_str());
            target->write(m_set_trx.shallow_clone(), mxs::Backend::IGNORE_RESPONSE);
        }
    }

    return rval;
}

bool RWSplitSession::prepare_target(RWBackend* target, route_target_t route_target)
{
    mxb_assert(target->in_use() || (!target->has_failed() && can_recover_servers()));
    return target->in_use() || prepare_connection(target);
}

void RWSplitSession::retry_query(GWBUF&& querybuf, int delay)
{
    mxb_assert(querybuf);

    // Route the query again later
    m_pSession->delay_routing(
        this, GWBUF(), std::chrono::seconds(delay), [this](GWBUF&&){
        mxb_assert(!m_pending_retries.empty());
        GWBUF buffer = std::move(m_pending_retries.front());
        m_pending_retries.erase(m_pending_retries.begin());

        if (m_canceled_retries > 0)
        {
            --m_canceled_retries;
            MXB_INFO("Discarding retried query: %s", get_sql_string(buffer).c_str());

            mxb_assert_message(replaying_trx(),
                               "Only transaction replay should cause retried queries to be discarded");
            return true;
        }


        return route_query(std::move(buffer));
    });

    ++m_retry_duration;

    m_pending_retries.push_back(std::move(querybuf));
}

bool RWSplitSession::have_connected_slaves() const
{
    return std::any_of(m_raw_backends.begin(), m_raw_backends.end(), [](auto b) {
        return b->is_slave() && b->in_use();
    });
}

bool RWSplitSession::should_try_trx_on_slave(route_target_t route_target) const
{
    return m_config->optimistic_trx                 // Optimistic transactions are enabled
           && !is_locked_to_master()                // Not locked to master
           && m_state == ROUTING                    // In normal routing mode
           && TARGET_IS_MASTER(route_target)        // The target type is master
           && have_connected_slaves()               // At least one connected slave
           && route_info().is_trx_still_read_only();// The start of the transaction is a read-only statement
}

void RWSplitSession::track_optimistic_trx(GWBUF& buffer, const RoutingPlan& res)
{
    if (res.type == RoutingPlan::Type::OTRX_START)
    {
        mxb_assert(res.route_target == TARGET_SLAVE);
        m_state = OTRX_STARTING;
    }
    else if (res.type == RoutingPlan::Type::OTRX_END)
    {
        mxb_assert(res.route_target == TARGET_LAST_USED);

        if (trx_is_ending())
        {
            m_state = ROUTING;
        }
        else if (!route_info().is_trx_still_read_only())
        {
            // Not a plain SELECT, roll it back on the slave and start it on the master
            MXB_INFO("Rolling back current optimistic transaction");

            /**
             * Store the actual statement we were attempting to execute and
             * replace it with a ROLLBACK. The storing of the statement is
             * done here to avoid storage of the ROLLBACK.
             */
            m_current_query.buffer = std::exchange(buffer, mariadb::create_query("ROLLBACK"));
            m_state = OTRX_ROLLBACK;
        }
    }
    else if (m_state == OTRX_STARTING)
    {
        mxb_assert(res.route_target == TARGET_LAST_USED);
        m_state = OTRX_ACTIVE;
    }
}

/**
 * Route query to all backends
 *
 * @param querybuf Query to route
 *
 * @return True if routing was successful
 */
bool RWSplitSession::handle_target_is_all(GWBUF&& buffer, const RoutingPlan& res)
{
    const RouteInfo& info = route_info();
    bool result = false;

    if (route_info().multi_part_packet())
    {
        continue_large_session_write(std::move(buffer), info.type_mask());
        result = true;
    }
    else if (route_session_write(std::move(buffer), info.command(), info.type_mask(), res))
    {
        result = true;
    }

    return result;
}

bool RWSplitSession::handle_routing_failure(GWBUF&& buffer, const RoutingPlan& res)
{
    bool ok = true;
    auto old_wait_gtid = m_wait_gtid;

    if (m_wait_gtid == READING_GTID)
    {
        mxb_assert(get_sql(buffer) == "SELECT @@gtid_current_pos");
        buffer = reset_gtid_probe();
    }

    if (std::all_of(m_raw_backends.begin(), m_raw_backends.end(), std::mem_fn(&mxs::RWBackend::has_failed)))
    {
        MXB_ERROR("All backends are permanently unusable for %s (%s: %s), closing connection.\n%s",
                  route_target_to_string(res.route_target), mariadb::cmd_to_string(buffer.data()[4]),
                  get_sql_string(buffer).c_str(), get_verbose_status().c_str());
        ok = false;
    }
    else if (replaying_trx())
    {
        ok = restart_trx_replay();
    }
    else if (should_migrate_trx() || (trx_is_open() && old_wait_gtid == READING_GTID))
    {
        // If the connection to the previous transaction target is still open, we must close it to prevent the
        // transaction from being accidentally committed whenever a new transaction is started on it.
        discard_connection(m_trx.target(), "Closed due to transaction migration");

        ok = start_trx_migration(std::move(buffer));
    }
    else if (can_retry_query())
    {
        MXB_INFO("Delaying routing: %s", get_sql_string(buffer).c_str());
        retry_query(std::move(buffer));
    }
    else if (m_config->master_failure_mode == RW_ERROR_ON_WRITE)
    {
        MXB_INFO("Sending read-only error, no valid target found for %s",
                 route_target_to_string(res.route_target));
        set_response(mariadb::create_error_packet(1, ER_OPTION_PREVENTS_STATEMENT, "HY000",
                                                  "The MariaDB server is running with the --read-only"
                                                  " option so it cannot execute this statement"));
        discard_connection(m_current_master, "The original primary is not available");
    }
    else if (res.route_target == TARGET_MASTER
             && (!m_config->delayed_retry || m_retry_duration >= m_config->delayed_retry_timeout.count()))
    {
        // Cannot retry the query, log a message that routing has failed
        log_master_routing_failure(res.target != nullptr, m_current_master, res.target);
        ok = false;
    }

    else
    {
        MXB_ERROR("Could not find valid server for target type %s (%s: %s), closing connection. %s",
                  route_target_to_string(res.route_target), mariadb::cmd_to_string(buffer.data()[4]),
                  get_sql_string(buffer).c_str(), get_verbose_status().c_str());
        ok = false;
    }

    return ok;
}

void RWSplitSession::send_readonly_error()
{
    int errnum = ER_OPTION_PREVENTS_STATEMENT;
    const char sqlstate[] = "HY000";
    const char errmsg[] = "The MariaDB server is running with the --read-only"
                          " option so it cannot execute this statement";

    mxs::ReplyRoute route;
    mxs::Reply reply;
    reply.set_error(errnum, sqlstate, sqlstate + sizeof(sqlstate) - 1, errmsg, errmsg + sizeof(errmsg) - 1);
    RouterSession::clientReply(mariadb::create_error_packet(1, errnum, sqlstate, errmsg), route, reply);
}

bool RWSplitSession::query_not_supported(const GWBUF& querybuf)
{
    bool unsupported = false;
    const RouteInfo& info = route_info();
    route_target_t route_target = info.target();
    GWBUF err;

    if (mxs_mysql_is_ps_command(info.command()) && info.stmt_id() == 0)
    {
        if (protocol_data().will_respond(querybuf))
        {
            // Unknown PS ID, can't route this query
            std::stringstream ss;
            ss << "Unknown prepared statement handler (" << extract_binary_ps_id(querybuf)
               << ") for " << mariadb::cmd_to_string(info.command()) << " given to MaxScale";
            err = mariadb::create_error_packet(1, ER_UNKNOWN_STMT_HANDLER, "HY000", ss.str().c_str());
            mxs::unexpected_situation(ss.str().c_str());
        }
        else
        {
            // The command doesn't expect a response which means we mustn't send one. Sending an unexpected
            // error will cause the client to go out of sync.
            unsupported = true;
        }
    }
    else if (TARGET_IS_ALL(route_target) && (TARGET_IS_MASTER(route_target) || TARGET_IS_SLAVE(route_target)))
    {
        // Conflicting routing targets. Return an error to the client.
        MXB_ERROR("Can't route %s '%s'. SELECT with session data modification is not "
                  "supported with `use_sql_variables_in=all`.",
                  mariadb::cmd_to_string(info.command()),
                  get_sql_string(querybuf).c_str());

        err = mariadb::create_error_packet(1, 1064, "42000", "Routing query to backend failed. "
                                                             "See the error log for further details.");
    }

    if (!err.empty())
    {
        set_response(std::move(err));
        unsupported = true;
    }

    return unsupported;
}

bool RWSplitSession::reuse_prepared_stmt(const GWBUF& buffer)
{
    const RouteInfo& info = route_info();

    if (info.command() == MXS_COM_STMT_PREPARE)
    {
        auto it = m_ps_cache.find(get_sql_string(buffer));

        if (it != m_ps_cache.end())
        {
            set_response(it->second.shallow_clone());
            return true;
        }
    }
    else if (info.command() == MXS_COM_STMT_CLOSE)
    {
        return true;
    }

    return false;
}

/**
 * Routes a buffer containing a single packet
 *
 * @param buffer The buffer to route
 *
 * @return True if routing succeed or if it failed due to unsupported query.
 *         false if backend failure was encountered.
 */
bool RWSplitSession::route_stmt(GWBUF&& buffer, const RoutingPlan& res)
{
    const RouteInfo& info = route_info();
    route_target_t route_target = info.target();
    mxb_assert_message(m_state != OTRX_ROLLBACK, "OTRX_ROLLBACK should never happen when routing queries");

    if (m_config->reuse_ps && reuse_prepared_stmt(buffer))
    {
        mxb::atomic::add(&m_router->stats().n_ps_reused, 1, mxb::atomic::RELAXED);
        return true;
    }

    if (query_not_supported(buffer))
    {
        return true;
    }
    else if (TARGET_IS_ALL(route_target))
    {
        return handle_target_is_all(std::move(buffer), res);
    }
    else
    {
        return route_single_stmt(std::move(buffer), res);
    }
}

bool RWSplitSession::route_single_stmt(GWBUF&& buffer, const RoutingPlan& res)
{
    bool ok = true;
    auto target = res.target;

    if (res.route_target == TARGET_MASTER && target != m_current_master)
    {
        if (should_replace_master(target))
        {
            MXB_INFO("Replacing old primary '%s' with new primary '%s'",
                     m_current_master ? m_current_master->name() : "<no previous master>",
                     target->name());
            replace_master(target);
        }
        else if (target)
        {
            MXB_INFO("Cannot replace old primary with '%s'", target->name());
            target = nullptr;
        }
    }

    if (target)
    {
        update_statistics(res);

        track_optimistic_trx(buffer, res);

        // We have a valid target, reset retry duration
        m_retry_duration = 0;

        if (!prepare_target(target, res.route_target))
        {
            // The connection to target was down and we failed to reconnect
            ok = false;
        }
        else
        {
            // If delayed query retry is enabled, we need to store the current statement
            bool store_stmt = m_state != OTRX_ROLLBACK
                && (m_config->delayed_retry
                    || (TARGET_IS_SLAVE(res.route_target) && m_config->retry_failed_reads));

            if (handle_got_target(std::move(buffer), target, store_stmt))
            {
                // Target server was found and is in the correct state. Store the original routing plan but
                // set the target as the actual target we routed it to.
                ok = true;
                m_prev_plan = res;
                m_prev_plan.target = target;

                mxb::atomic::add(&m_router->stats().n_queries, 1, mxb::atomic::RELAXED);
            }
        }
    }
    else
    {
        ok = handle_routing_failure(std::move(buffer), res);
    }

    return ok;
}

RWBackend* RWSplitSession::get_target(const GWBUF& buffer, route_target_t route_target)
{
    RWBackend* rval = nullptr;
    const RouteInfo& info = route_info();

    // We can't use a switch here as the route_target is a bitfield where multiple values are set at one time.
    // Mostly this happens when the type is TARGET_NAMED_SERVER and TARGET_SLAVE due to a routing hint.
    if (TARGET_IS_NAMED_SERVER(route_target) || TARGET_IS_RLAG_MAX(route_target))
    {
        // If transaction replay is enabled and a transaction is open, hints must be ignored. This prevents
        // them from overriding the transaction target which is what would otherwise happen and which causes
        // problems.
        if (m_config->transaction_replay && trx_is_open() && m_trx.target())
        {
            MXB_INFO("Transaction replay is enabled, ignoring routing hint while inside a transaction.");
        }
        else
        {
            return handle_hinted_target(buffer, route_target);
        }
    }

    if (TARGET_IS_LAST_USED(route_target))
    {
        rval = get_last_used_backend();
    }
    else if (TARGET_IS_SLAVE(route_target))
    {
        rval = handle_slave_is_target(info.command(), info.stmt_id());
    }
    else if (TARGET_IS_MASTER(route_target))
    {
        rval = handle_master_is_target();
    }
    else
    {
        MXB_ERROR("Unexpected target type: %s", route_target_to_string(route_target));
        mxb_assert(!true);
    }

    return rval;
}

RWSplitSession::RoutingPlan RWSplitSession::resolve_route(const GWBUF& buffer, const RouteInfo& info)
{
    RoutingPlan rval;
    rval.route_target = info.target();

    if (m_config->max_slave_connections == 0 && rval.route_target == TARGET_SLAVE)
    {
        // With max_slave_connections=0, all reads can be treated as writes. This will correctly trigger a
        // master migration if the current one no longer qualifies for it and a replacement is available.
        // Otherwise, the migration would not happen until the next write.
        rval.route_target = TARGET_MASTER;
    }

    if (info.multi_part_packet())
    {
        /** We're processing a large query that's split across multiple packets.
         * Route it to the same backend where we routed the previous packet. */
        rval.route_target = TARGET_LAST_USED;
    }
    else if (trx_is_starting() && !trx_is_read_only() && should_try_trx_on_slave(rval.route_target))
    {
        // A normal transaction is starting and it qualifies for speculative routing
        rval.type = RoutingPlan::Type::OTRX_START;
        rval.route_target = TARGET_SLAVE;
    }
    else if (m_state == OTRX_STARTING || m_state == OTRX_ACTIVE)
    {
        if (trx_is_ending() || !info.is_trx_still_read_only())
        {
            rval.type = RoutingPlan::Type::OTRX_END;
        }

        rval.route_target = TARGET_LAST_USED;
    }

    if (rval.route_target != TARGET_ALL)
    {
        rval.target = get_target(buffer, rval.route_target);
    }

    return rval;
}

bool RWSplitSession::write_session_command(RWBackend* backend, GWBUF&& buffer, uint8_t cmd)
{
    bool ok = true;
    mxs::Backend::response_type type = mxs::Backend::NO_RESPONSE;

    if (protocol_data().will_respond(buffer))
    {
        type = backend == m_sescmd_replier ? mxs::Backend::EXPECT_RESPONSE : mxs::Backend::IGNORE_RESPONSE;
    }

    if (backend->write(std::move(buffer), type))
    {
        auto& stats = m_router->local_server_stats()[backend->target()];
        stats.inc_total();
        stats.inc_read();
        MXB_INFO("Route query to %s: %s", backend == m_current_master ? "primary" : "replica",
                 backend->name());
    }
    else
    {
        MXB_ERROR("Failed to execute session command in %s", backend->name());
        backend->close();

        if (m_config->master_failure_mode == RW_FAIL_INSTANTLY && backend == m_current_master)
        {
            ok = false;
        }
    }

    return ok;
}

/**
 * Execute in backends used by current router session.
 * Save session variable commands to router session property
 * struct. Thus, they can be replayed in backends which are
 * started and joined later.
 *
 * Suppress redundant OK packets sent by backends.
 *
 * The first OK packet is replied to the client.
 *
 * @param querybuf      GWBUF including the query to be routed
 * @param inst          Router instance
 * @param packet_type       Type of MySQL packet
 * @param qtype         Query type from query_classifier
 *
 * @return True if at least one backend is used and routing succeed to all
 * backends being used, otherwise false.
 *
 */
bool RWSplitSession::route_session_write(GWBUF&& buffer, uint8_t command, uint32_t type,
                                         const RoutingPlan& plan)
{
    MXB_INFO("Session write, routing to all servers.");
    bool ok = true;
    std::ostringstream error;

    if (!have_open_connections() || need_master_for_sescmd())
    {
        MXB_INFO("No connections available for session command");

        if (command == MXS_COM_QUIT)
        {
            // We have no open connections and opening one just to close it is pointless.
            MXB_INFO("Ignoring COM_QUIT");
            return true;
        }
        else if (can_recover_servers())
        {
            MXB_INFO("Attempting to create a connection");
            // No connections are open, create one and execute the session command on it
            create_one_connection_for_sescmd();
        }
    }

    // Pick a new replier for each new session command. This allows the source server to change over
    // the course of the session. The replier will usually be the current master server.
    m_sescmd_replier = nullptr;

    for (RWBackend* backend : m_raw_backends)
    {
        if (backend->in_use())
        {
            if (!m_sescmd_replier || backend == m_current_master)
            {
                // Return the result from this backend to the client
                m_sescmd_replier = backend;
            }
        }
    }

    if (trx_is_open() && m_trx.target() && m_trx.target()->in_use())
    {
        // A transaction is open on a backend, use it instead.
        m_sescmd_replier = m_trx.target();
    }

    if (m_sescmd_replier && need_master_for_sescmd())
    {
        MXB_INFO("Cannot use '%s' for session command: transaction is open.", m_sescmd_replier->name());
        m_sescmd_replier = nullptr;
    }

    if (m_sescmd_replier)
    {
        for (RWBackend* backend : m_raw_backends)
        {
            if (backend->in_use() && !write_session_command(backend, buffer.shallow_clone(), command))
            {
                ok = false;
            }
        }

        if (ok)
        {
            if (command == MXS_COM_STMT_CLOSE)
            {
                // Remove the command from the PS mapping
                m_qc.ps_erase(&buffer);

                auto stmt_id = route_info().stmt_id();
                auto it = std::find(m_exec_map.begin(), m_exec_map.end(), ExecInfo {stmt_id});

                if (it != m_exec_map.end())
                {
                    m_exec_map.erase(it);
                }
            }
            else if (Parser::type_mask_contains(type, mxs::sql::TYPE_PREPARE_NAMED_STMT)
                     || Parser::type_mask_contains(type, mxs::sql::TYPE_PREPARE_STMT))
            {
                mxb_assert(buffer.id() != 0
                           || Parser::type_mask_contains(type, mxs::sql::TYPE_PREPARE_NAMED_STMT));
                m_qc.ps_store(&buffer, buffer.id());
            }
            else if (Parser::type_mask_contains(type, mxs::sql::TYPE_DEALLOC_PREPARE))
            {
                mxb_assert(!mxs_mysql_is_ps_command(route_info().command()));
                m_qc.ps_erase(&buffer);
            }

            // TODO: Fix this
            // m_router->update_max_sescmd_sz(protocol_data().history().size());

            m_current_query.buffer = std::move(buffer);

            // Session command routed, reset retry duration
            m_retry_duration = 0;

            m_prev_plan = plan;

            mxb::atomic::add(&m_router->stats().n_all, 1, mxb::atomic::RELAXED);
            mxb::atomic::add(&m_router->stats().n_queries, 1, mxb::atomic::RELAXED);

            if (protocol_data().will_respond(m_current_query.buffer))
            {
                m_expected_responses++;
                mxb_assert(m_expected_responses == 1);
                MXB_INFO("Will return response from '%s' to the client", m_sescmd_replier->name());
            }

            if (trx_is_open() && !m_trx.target())
            {
                m_trx.set_target(m_sescmd_replier);
            }
            else
            {
                mxb_assert_message(!trx_is_open() || m_trx.target() == m_sescmd_replier,
                                   "Trx target is %s when m_sescmd_replier is %s while trx is open",
                                   m_trx.target() ? m_trx.target()->name() : "nullptr",
                                   m_sescmd_replier->name());
            }
        }
        else
        {
            error << "Could not route session command "
                  << "(" << mariadb::cmd_to_string(command) << ": "
                  << get_sql(buffer) << ").";
        }
    }
    else
    {
        error << "No valid candidates for session command "
              << "(" << mariadb::cmd_to_string(command) << ": "
              << get_sql(buffer) << ").";
        ok = false;
    }

    if (!ok)
    {
        if (replaying_trx())
        {
            ok = restart_trx_replay();
        }
        else if (can_retry_query())
        {
            // TODO: This unnecessarily requires delayed_retry when it should work without it
            MXB_INFO("Delaying routing: %s", get_sql_string(buffer).c_str());
            retry_query(std::move(buffer));
            ok = true;
        }
        else
        {
            if (m_retry_duration >= m_config->delayed_retry_timeout.count())
            {
                error << " Retry took too long (" << m_retry_duration << " seconds).";
            }

            error << " Connection status: " << get_verbose_status();

            MXB_ERROR("%s", error.str().c_str());
        }
    }

    return ok;
}

RWBackend* RWSplitSession::get_hinted_backend(const char* name)
{
    RWBackend* rval = nullptr;

    for (auto backend : m_raw_backends)
    {
        /** The server must be a valid slave, relay server, or master */
        if ((backend->in_use() || (can_recover_servers() && backend->can_connect()))
            && strcasecmp(name, backend->name()) == 0)
        {
            rval = backend;
            break;
        }
    }

    return rval;
}

RWBackend* RWSplitSession::get_master_backend()
{
    RWBackend* rval = nullptr;

    if (RWBackend* master = get_root_master())
    {
        if (is_valid_for_master(master))
        {
            rval = master;
        }
    }

    return rval;
}

RWBackend* RWSplitSession::get_last_used_backend()
{
    return m_prev_plan.target ? m_prev_plan.target : get_master_backend();
}

/**
 * Provide the router with a reference to a suitable backend
 *
 * @param btype    Backend type
 * @param name     Name of the requested backend. May be nullptr if any name is accepted.
 * @param max_rlag Maximum replication lag
 *
 * @return The backend if one was found
 */
RWBackend* RWSplitSession::get_target_backend(backend_type_t btype, const char* name, int max_rlag)
{
    RWBackend* rval = nullptr;

    if (trx_is_open() && m_trx.target() && m_wait_gtid != READING_GTID)
    {
        // A transaction that has an existing target. Continue using it as long as it remains valid.
        if (trx_target_still_valid())
        {
            rval = m_trx.target();
        }
    }
    else if (name)
    {
        // Choose backend by name from a hint
        rval = get_hinted_backend(name);
    }
    else if (btype == BE_SLAVE)
    {
        rval = get_slave_backend(max_rlag);
    }
    else if (btype == BE_MASTER)
    {
        rval = get_master_backend();
    }

    return rval;
}

/**
 * @brief Get the maximum replication lag for this router
 *
 * @param   rses    Router client session
 * @return  Replication lag from configuration or very large number
 */
int RWSplitSession::get_max_replication_lag()
{
    int conf_max_rlag = mxs::Target::RLAG_UNDEFINED;

    /** if there is no configured value, then longest possible int is used */
    if (m_config->max_replication_lag.count() > 0)
    {
        conf_max_rlag = m_config->max_replication_lag.count();
    }

    return conf_max_rlag;
}

/**
 * @brief Handle hinted target query
 *
 * One of the possible types of handling required when a request is routed
 *
 *  @param ses          Router session
 *  @param querybuf     Buffer containing query to be routed
 *  @param route_target Target for the query
 *  @param target_dcb   DCB for the target server
 *
 *  @return bool - true if succeeded, false otherwise
 */
RWBackend* RWSplitSession::handle_hinted_target(const GWBUF& querybuf, route_target_t route_target)
{
    int config_max_rlag = get_max_replication_lag();    // From router configuration.
    RWBackend* target = nullptr;

    for (const Hint& hint : querybuf.hints())
    {
        if (hint.type == Hint::Type::ROUTE_TO_NAMED_SERVER)
        {
            // Set the name of searched backend server.
            const char* named_server = hint.data.c_str();
            target = get_target_backend(BE_UNDEFINED, named_server, config_max_rlag);
            MXB_INFO("Hint: route to server '%s', %s.", named_server,
                     target ? "found target" : "target not valid");
        }
        else if (hint.type == Hint::Type::PARAMETER
                 && (mxb::sv_case_eq(hint.data, "max_replication_lag")
                     || mxb::sv_case_eq(hint.data, "max_slave_replication_lag")))
        {
            auto hint_max_rlag = strtol(hint.value.c_str(), nullptr, 10);
            if (hint_max_rlag > 0)
            {
                target = get_target_backend(BE_SLAVE, nullptr, hint_max_rlag);
                MXB_INFO("Hint: %s=%s, %s.", hint.data.c_str(), hint.value.c_str(),
                         target ? "found target" : "target not valid");
            }
            else
            {
                MXB_INFO("Ignoring invalid hint value: %s", hint.value.c_str());
            }
        }

        if (target)
        {
            break;
        }
    }

    if (!target)
    {
        // If no target so far, pick any available. TODO: should this be error instead?
        // Erroring here is more appropriate when namedserverfilter allows setting multiple target types
        // e.g. target=server1,->slave

        backend_type_t btype = route_target & TARGET_SLAVE ? BE_SLAVE : BE_MASTER;
        target = get_target_backend(btype, nullptr, config_max_rlag);
    }
    return target;
}

/**
 * Handle slave target type
 *
 * @param cmd     Command being executed
 * @param stmt_id Prepared statement ID
 *
 * @return The target backend if one was found
 */
RWBackend* RWSplitSession::handle_slave_is_target(uint8_t cmd, uint32_t stmt_id)
{
    int rlag_max = get_max_replication_lag();
    RWBackend* target = nullptr;

    if (route_info().is_ps_continuation())
    {
        auto it = std::find(m_exec_map.begin(), m_exec_map.end(), ExecInfo {stmt_id});

        if (it != m_exec_map.end() && it->target)
        {
            auto prev_target = it->target;

            if (prev_target->in_use())
            {
                target = prev_target;
                MXB_INFO("%s on %s", mariadb::cmd_to_string(cmd), target->name());
            }
            else
            {
                MXB_ERROR("Old COM_STMT_EXECUTE target %s not in use, cannot "
                          "proceed with %s", prev_target->name(), mariadb::cmd_to_string(cmd));
            }
        }
        else
        {
            MXB_WARNING("Unknown statement ID %u used in %s", stmt_id, mariadb::cmd_to_string(cmd));
        }
    }
    else
    {
        target = get_target_backend(BE_SLAVE, nullptr, rlag_max);
    }

    if (!target)
    {
        MXB_INFO("Was supposed to route to replica but finding suitable one failed.");
    }

    return target;
}

/**
 * @brief Log master write failure
 */
void RWSplitSession::log_master_routing_failure(bool found,
                                                RWBackend* old_master,
                                                RWBackend* curr_master)
{
    std::string errmsg;

    if (m_config->delayed_retry && m_retry_duration >= m_config->delayed_retry_timeout.count())
    {
        errmsg = get_delayed_retry_failure_reason();
    }
    else if (!found)
    {
        errmsg = mxb::string_printf("Could not find a valid master connection");
    }
    else if (old_master && curr_master && old_master->in_use())
    {
        /** We found a master but it's not the same connection */
        mxb_assert(old_master != curr_master);
        errmsg = mxb::string_printf(
            "Master server changed from '%s' to '%s'",
            old_master->name(),
            curr_master->name());
    }
    else if (old_master && old_master->in_use())
    {
        // TODO: Figure out if this is an impossible situation
        mxb_assert(!curr_master);
        /** We have an original master connection but we couldn't find it */
        errmsg = mxb::string_printf(
            "The connection to primary server '%s' is not available",
            old_master->name());
    }
    else
    {
        /** We never had a master connection, the session must be in read-only mode */
        if (m_config->master_failure_mode != RW_FAIL_INSTANTLY)
        {
            errmsg = mxb::string_printf(
                "Session is in read-only mode because it was created "
                "when no primary was available");
        }
        else
        {
            mxb_assert(old_master && !old_master->in_use());
            errmsg = "Was supposed to route to primary but the primary connection is closed";
        }
    }

    MXB_WARNING("Write query received from %s@%s. %s. Closing client connection.",
                m_pSession->user().c_str(),
                m_pSession->client_remote().c_str(),
                errmsg.c_str());
}

bool RWSplitSession::should_replace_master(RWBackend* target)
{
    return m_config->master_reconnection
           &&   // We have a target server and it's not the current master
           target && target != m_current_master
           &&   // We are not inside a transaction (also checks for autocommit=1)
           (!trx_is_open() || trx_is_starting() || (replaying_trx() && !m_trx.target()))
           &&   // We are not locked to the old master
           !is_locked_to_master()
           &&   // The server is actually labeled as a master
           target->is_master();
}

void RWSplitSession::discard_connection(mxs::RWBackend* target, const std::string& error)
{
    if (target && target->in_use())
    {
        MXB_INFO("Discarding connection to '%s': %s", target->name(), error.c_str());
        target->close();

        if (target == m_current_master)
        {
            m_qc.master_replaced();
        }
    }
}

void RWSplitSession::replace_master(RWBackend* target)
{
    discard_connection(m_current_master, "The original primary is not available");
    m_current_master = target;
}

bool RWSplitSession::trx_target_still_valid() const
{
    bool ok = false;

    if (auto target = m_trx.target(); target != nullptr && target->in_use())
    {
        ok = target->is_master() || (trx_is_read_only() && target->is_slave());
    }

    return ok;
}

bool RWSplitSession::should_migrate_trx() const
{
    bool migrate = false;

    if (m_config->transaction_replay
        && !replaying_trx()     // Transaction replay is not active
        && trx_is_open()        // We have an open transaction
        && m_can_replay_trx)    // The transaction can be replayed
    {
        if (!trx_target_still_valid())
        {
            migrate = true;
        }
    }

    return migrate;
}

bool RWSplitSession::start_trx_migration(GWBUF&& querybuf)
{
    if (mxb_log_should_log(LOG_INFO) && m_trx.target())
    {
        MXB_INFO("Transaction target '%s' is no longer valid, replaying transaction", m_trx.target()->name());
    }

    /**
     * Stash the current query so that the transaction replay treats
     * it as if the query was interrupted.
     */
    m_current_query.buffer = std::move(querybuf);

    /**
     * After the transaction replay has been started, the rest of
     * the query processing needs to be skipped. This is done to avoid
     * the error logging done when no valid target is found for a query
     * as well as to prevent retrying of queries in the wrong order.
     */
    return start_trx_replay();
}

bool RWSplitSession::restart_trx_replay()
{
    mxb_assert(m_config->transaction_replay);
    MXB_INFO("Restarting transaction replay, no valid targets found. Last target was: %s",
             m_trx.target() ? m_trx.target()->name() : "<none>");

    // During transaction replay, whenever the routing of a query fails due to
    // a lack of valid targets, close the connection to the transaction target
    // and replay the whole transaction from the beginning. This will make sure
    // that any partially replayed transactions are closed as soon as possible.
    // Leaving the connection open causes any transactions that it has to also remain
    // open on the replay target which may prevent other operations from proceeding.
    //
    // In particular, this would happen if mariadbmon did a switchover while readwritesplit
    // was replaying a transaction on the server being demoted. As the switchover removes
    // the Master label from the demotion target before putting it into read-only mode,
    // a deadlock occurs where the monitor waits for the `SET GLOBAL read_only=1` to
    // complete which is blocked by the earlier transaction that readwritesplit has open
    // that in turn is waiting for a server with a Master label to appear. This cycle is
    // currently broken by the discard_connect() call that closes the connection which
    // rolls back all transactions.
    discard_connection(m_trx.target(), "Closed due to restart of transaction replay");
    return start_trx_replay();
}

/**
 * @brief Handle master is the target
 *
 * One of the possible types of handling required when a request is routed
 *
 *  @param inst         Router instance
 *  @param ses          Router session
 *  @param target_dcb   DCB for the target server
 *
 *  @return bool - true if succeeded, false otherwise
 */
RWBackend* RWSplitSession::handle_master_is_target()
{
    return get_target_backend(BE_MASTER, nullptr, mxs::Target::RLAG_UNDEFINED);
}

/**
 * @brief Handle writing to a target server
 *
 *  @return True on success
 */
bool RWSplitSession::handle_got_target(GWBUF&& buffer, RWBackend* target, bool store)
{
    mxb_assert_message(target->in_use(), "Target must be in use before routing to it");

    MXB_INFO("Route query to %s: %s <", target == m_current_master ? "primary" : "replica", target->name());

    uint8_t cmd = mxs_mysql_get_command(buffer);
    bool will_respond = parser().command_will_respond(cmd);

    bool attempting_causal_read = false;

    if (route_info().multi_part_packet() || route_info().load_data_active())
    {
        // Never store multi-packet queries or data sent during LOAD DATA LOCAL INFILE
        store = false;
        will_respond = false;
    }
    else if (!is_locked_to_master())
    {
        mxb_assert(!mxs_mysql_is_ps_command(cmd)
                   || extract_binary_ps_id(buffer) == route_info().stmt_id()
                   || extract_binary_ps_id(buffer) == MARIADB_PS_DIRECT_EXEC_ID);

        // Attempt a causal read only when the query is routed to a slave
        attempting_causal_read = target->is_slave() && should_do_causal_read();

        if (cmd == MXS_COM_QUERY && attempting_causal_read)
        {
            add_prefix_wait_gtid(buffer);
            store = false;      // The storage for causal reads is done inside add_prefix_wait_gtid
        }

        if (m_wait_gtid == GTID_READ_DONE)
        {
            // GTID sync done but causal read wasn't started because the conditions weren't met. Go back to
            // the default state since this now a normal read.
            m_wait_gtid = NONE;
        }
        else if (m_wait_gtid == READING_GTID)
        {
            store = false;      // This is the GTID sync query, don't store it
        }

        if (target->is_slave() && (cmd == MXS_COM_QUERY || cmd == MXS_COM_STMT_EXECUTE))
        {
            target->select_started();
        }

        if (cmd == MXS_COM_STMT_EXECUTE || cmd == MXS_COM_STMT_SEND_LONG_DATA)
        {
            // Track the targets of the COM_STMT_EXECUTE statements. This information is used to route all
            // COM_STMT_FETCH commands to the same server where the COM_STMT_EXECUTE was done.
            auto stmt_id = route_info().stmt_id();
            auto it = std::find(m_exec_map.begin(), m_exec_map.end(), ExecInfo {stmt_id});

            if (it == m_exec_map.end())
            {
                m_exec_map.emplace_back(stmt_id, target);
            }
            else
            {
                it->target = target;
            }

            MXB_INFO("%s on %s", mariadb::cmd_to_string(cmd), target->name());
        }
    }
    else if (cmd == MXS_COM_STMT_PREPARE)
    {
        // This is here to avoid a debug assertion in the ps_store_response call that is hit when we're locked
        // to the master due to strict_multi_stmt or strict_sp_calls and the user executes a prepared
        // statement. The previous PS ID is tracked in ps_store and asserted to be the same in
        // ps_store_result.
        m_qc.ps_store(&buffer, buffer.id());
    }

    if (store)
    {
        m_current_query.buffer = buffer.shallow_clone();

        if (m_config->transaction_replay && m_config->trx_retry_safe_commit
            && parser().type_mask_contains(route_info().type_mask(), mxs::sql::TYPE_COMMIT))
        {
            MXB_INFO("Transaction is about to commit, skipping replay if it fails.");
            m_can_replay_trx = false;
        }
    }

    mxs::Backend::response_type response = mxs::Backend::NO_RESPONSE;

    if (will_respond)
    {
        mxb_assert(!route_info().multi_part_packet());

        ++m_expected_responses;     // The server will reply to this command
        response = mxs::Backend::EXPECT_RESPONSE;
    }

    if (Parser::type_mask_contains(route_info().type_mask(), mxs::sql::TYPE_NEXT_TRX))
    {
        m_set_trx = buffer.shallow_clone();
    }

    if (trx_is_open())
    {
        if (m_wait_gtid == READING_GTID)
        {
            // Ignore transaction target if a sync query is in progress. This prevents the transaction from
            // being assigned based on the target of the sync query which would end up causing all read-only
            // transactions to be routed to the master.
            MXB_INFO("Doing GTID sync on '%s' while transaction is open, transaction target is '%s'",
                     target->name(), m_trx.target() ? m_trx.target()->name() : "<none>");
        }
        else if (!m_trx.target())
        {
            MXB_INFO("Transaction starting on '%s'", target->name());
            m_trx.set_target(target);
        }
        else if (trx_is_starting())
        {
            MXB_INFO("Transaction did not finish on '%s' before a new one started on '%s'",
                     m_trx.target()->name(), target->name());
            m_trx.close();
            m_trx.set_target(target);
        }
        else
        {
            mxb_assert(m_trx.target() == target);
        }
    }

    if (attempting_causal_read && cmd == MXS_COM_STMT_EXECUTE)
    {
        send_sync_query(target);
    }

    return target->write(std::move(buffer), response);
}
